package org.vstu.compprehension.models.businesslogic;

import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.Data;
import lombok.NoArgsConstructor;
import org.jetbrains.annotations.Nullable;
import org.vstu.compprehension.models.entities.EnumData.SearchDirections;
import org.vstu.compprehension.models.entities.QuestionRequestLogEntity;

import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

@Builder(toBuilder = true)
@NoArgsConstructor
@AllArgsConstructor
@Data
public class QuestionRequest {
    private List<Concept> deniedConcepts;
    private List<Concept> targetConcepts;
    /** Question Storage can treat this as "optional targets" or "preferred" */
    private List<Concept> allowedConcepts;
    private List<Law> targetLaws;
    private List<Law> deniedLaws;
    private List<Law> allowedLaws;
    private List<Tag> targetTags;

    private @Nullable List<String> deniedQuestionNames;
    private @Nullable List<Integer> deniedQuestionTemplateIds;
    private @Nullable List<Integer> deniedQuestionMetaIds; // same as deniedQuestionNames but using ids instead of names

    /**
     * Условная единица, показывающая долго или быстро решается вопрос
     * 1 - быстро, 10 - очень долго
     */
    private int solvingDuration;
    /**
     * Сложность задания [0..1]
     */
    private float complexity;

    /** минимум шагов в решении */
    private int stepsMin;
    /** максимум шагов в решении */
    private int stepsMax;

    /**
     * Направление поиска по сложности
     */
    private SearchDirections complexitySearchDirection;

    /**
     * Направление поиска по сложности
     */
    private SearchDirections lawsSearchDirection;

    /**
     * Probability of choosing an auto-generated question
     * Вероятность выбора авто-сгенерированного вопроса
     * range: [0..1]
     * 0: always pick manual question
     * 1: always pick auto-generated question
     */
    @Builder.Default
    private double chanceToPickAutogeneratedQuestion = 1;

    private String domainShortname;

    private @Nullable Long exerciseAttemptId;

    public QuestionRequestLogEntity getLogEntity(int questionsFound) {
        return QuestionRequestLogEntity.builder()
                // .id(0L)
                .exerciseAttemptId(exerciseAttemptId)
                .domainShortname(domainShortname)
                .targetConceptNames(targetConcepts.stream().map(Concept::getName).collect(Collectors.toList()))
                .deniedConceptNames(deniedConcepts.stream().map(Concept::getName).collect(Collectors.toList()))
                .allowedConceptNames(allowedConcepts.stream().map(Concept::getName).collect(Collectors.toList()))
                .targetLawNames(targetLaws.stream().map(Law::getName).collect(Collectors.toList()))
                .deniedLawNames(deniedLaws.stream().map(Law::getName).collect(Collectors.toList()))
                .allowedLawNames(allowedLaws.stream().map(Law::getName).collect(Collectors.toList()))
                .targetTags(targetTags.stream().map(Tag::getName).collect(Collectors.toList()))
                .conceptsTargetedBitmask(Concept.combineToBitmask(targetConcepts))
                .conceptsDeniedBitmask(Concept.combineToBitmask(deniedConcepts))
                .lawsTargetedBitmask(Law.combineToBitmask(targetLaws))
                .lawsDeniedBitmask(Law.combineToBitmask(deniedLaws))
                .targetTagsBitmask(Tag.combineToBitmask(targetTags))
                .deniedQuestionNames(deniedQuestionNames)
                .deniedQuestionTemplateIds(deniedQuestionTemplateIds)
                .deniedQuestionMetaIds(deniedQuestionMetaIds)
                .solvingDuration(solvingDuration)
                .complexity(complexity)
                .stepsMin(stepsMin)
                .stepsMax(stepsMax)
                .complexitySearchDirection(complexitySearchDirection)
                .lawsSearchDirection(lawsSearchDirection)
                .chanceToPickAutogeneratedQuestion(chanceToPickAutogeneratedQuestion)
                .createdDate(new Date())
                .foundCount(questionsFound)
        .build();
    }
}

